﻿using System;
using System.Collections;
using System.Threading;
using Microsoft.SPOT;
using Microsoft.SPOT.Presentation;
using Microsoft.SPOT.Presentation.Controls;
using Microsoft.SPOT.Presentation.Media;
using Microsoft.SPOT.Presentation.Shapes;
using Microsoft.SPOT.Touch;

using Microsoft.SPOT.Net.NetworkInformation;
using Microsoft.SPOT.Time;

using Gadgeteer.Networking;
using GT = Gadgeteer;
using GTM = Gadgeteer.Modules;
using Gadgeteer.Modules.GHIElectronics;

using Microsoft.WindowsAzure.MobileServices;
using Microsoft.Azure.Zumo.MicroFramework.Core;

namespace Accel
{
    public partial class Program
    {
        private const bool debug = true; // control debug display

        private MobileServiceClient mobileService;
        private static readonly string WAMSAppKey = "WAMS-APP-KEY";
        private static readonly string WAMSUrl = "http://iot-kit.azure-mobile.net/";
        private static readonly string TableName = "accel";
        private static readonly Guid SensorDeviceId = new Guid(
            // {ECA12D7A-D9CF-4646-B2BB-6CEA16741C04}
            0xeca12d7a, 0xd9cf, 0x4646, 0xb2, 0xbb, 0x6c, 0xea, 0x16, 0x74, 0x1c, 0x4);

        private string ipAddress = "";
        private bool isNetworkConnected = false;
        private bool isAzureOK = false;

        private const int timeSec = 10;
        private const int timerMS = timeSec * 1000;
        private const int timerAccelG248 = 479; // 479ms
        private static byte[] TimeServerIPAddress = new byte[] { 133, 243, 238, 163 }; //ntp-b2.nict.jp

        private AccelG248.Acceleration nowAccel;
        private AccelG248.Acceleration lastAccel;
        private AccelG248.Acceleration diffAccel;
        private AccelG248.Acceleration maxDiffAccel;
        private AccelG248.Acceleration nowG248;
        private AccelG248.Acceleration lastG248;
        private AccelG248.Acceleration diffG248;
        private AccelG248.Acceleration maxDiffG248;
        private const double accelThreshold = 0.10;
        private const double g248Threshold = 0.10;

        // This method is run when the mainboard is powered up or reset.   
        void ProgramStarted()
        {
            // Use Debug.Print to show messages in Visual Studio's "Output" window during debugging.
            Debug.Print("Program Started, interval = " + timeSec + " secs");

            SetupNetwork();
            if (isNetworkConnected)
            {
                SyncTimeService(); // Get time of day
                Thread.Sleep(500); // Wait for time event
                if (timeSynced)
                {
                    Debug.Print("Time synced: " + DateTime.Now);
                }
                else
                {
                    Debug.Print("Warning! Time is not synced");
                    multicolorLED.FadeRepeatedly(GT.Color.Purple);
                }
            }
            else
            {
                Debug.Print("No network connected. Running on local mode.");
            }

            SetupChecker();
            if (isNetworkConnected)
            {
                SetupMobileService();
                SetupAccelG248();
                SetupAccelerometer();
            }
            else
            {
                multicolorLED.TurnWhite(); // Network error
            }
        }

        private void SetupNetwork()
        {
            string saveIpAddress = "";
            string saveSubnetMask = "";
            string saveGatewayAddress = "";
            string[] saveDnsAddresses = null;
            GHI.Networking.EthernetBuiltIn ei = ethernetJ11D.NetworkInterface;
            NetworkInterface en = ethernetJ11D.NetworkSettings;

            Thread.Sleep(500);
            if (!ei.Opened)
            {
                Debug.Print("now open J11D");
                ei.Open();
            }
            if (ei.IsDhcpEnabled)
            {
                Debug.Print("now enable DHCP");
                ei.EnableDhcp();
            }
            else
            {
                // Save Flash ROM settings
                char[] caIpAddress = en.IPAddress.ToCharArray();
                saveIpAddress = new string(caIpAddress);

                char[] caSubnetMask = en.SubnetMask.ToCharArray();
                saveSubnetMask = new string(caSubnetMask);

                char[] caGatewayAddress = en.GatewayAddress.ToCharArray();
                saveGatewayAddress = new string(caGatewayAddress);

                int numOfDns = en.DnsAddresses.Length;
                saveDnsAddresses = new string[numOfDns];
                if (numOfDns > 0)
                {
                    char[] dnsAddress0 = en.DnsAddresses[0].ToCharArray();
                    saveDnsAddresses[0] = new string(dnsAddress0);
                }
                if (numOfDns > 1)
                {
                    char[] dnsAddress1 = en.DnsAddresses[1].ToCharArray();
                    saveDnsAddresses[1] = new string(dnsAddress1);
                }
            }

            PrintNetworkInfo("Start settings");
            foreach (var ni in NetworkInterface.GetAllNetworkInterfaces())
            {
                if (ni.NetworkInterfaceType == NetworkInterfaceType.Ethernet)
                {
                    if (ni.IsDhcpEnabled)
                    {
                        ipAddress = ni.IPAddress;
                        if (ipAddress != null && ipAddress != "0.0.0.0")
                        {
                            break;
                        }
                        else
                        {
                            ni.RenewDhcpLease();
                            Thread.Sleep(500);
                            ipAddress = ni.IPAddress;
                        }
                    }
                    else
                    {
                        ni.EnableDhcp();
                        Thread.Sleep(500);
                        ni.EnableStaticIP(saveIpAddress, saveSubnetMask, saveGatewayAddress);
                        ni.EnableStaticDns(saveDnsAddresses);
                        ipAddress = ni.IPAddress;
                    }
                    break;
                }
            }
            isNetworkConnected = ipAddress != null && ipAddress != "0.0.0.0";
            PrintNetworkInfo("Completed");
        }

        void PrintNetworkInfo(string arg)
        {
            NetworkInterface en = ethernetJ11D.NetworkSettings;
            const string line = "-----------------";
            Debug.Print(line + " " + arg + " " + line);
            Debug.Print("IP Address:   " + en.IPAddress);
            Debug.Print("DHCP Enabled: " + en.IsDhcpEnabled);
            Debug.Print("Subnet Mask:  " + en.SubnetMask);
            Debug.Print("Gateway:      " + en.GatewayAddress);
            Debug.Print(line + line + line);
        }

        private void SyncTimeService()
        {
            TimeService.SystemTimeChanged += TimeService_SystemTimeChanged;
            TimeServiceStatus status = TimeService.UpdateNow(TimeServerIPAddress, 10);
            TimeService.SetTimeZoneOffset(540); // JST time origin 9h
        }

        bool timeSynced = false;
        void TimeService_SystemTimeChanged(object sender, SystemTimeChangedEventArgs e)
        {
            lock (this)
            {
                timeSynced = true;
            }
        }

        private void SetupChecker()
        {
            GT.Timer timer = new GT.Timer(timerMS);
            timer.Tick += MeasurementChecker;
            timer.Behavior = GT.Timer.BehaviorType.RunContinuously;
            timer.Start();
        }
        private void SetupAccelG248()
        {
            maxDiffG248.X = diffG248.X = lastG248.X = nowG248.X = 0.0;
            maxDiffG248.Y = diffG248.Y = lastG248.Y = nowG248.Y = 0.0;
            maxDiffG248.Z = diffG248.Z = lastG248.Z = nowG248.Y = 0.0;

            GT.Timer timer = new GT.Timer(timerAccelG248);
            timer.Tick += AccelG248_Tick;
            timer.Behavior = GT.Timer.BehaviorType.RunContinuously;
            timer.Start();
        }

        private void SetupAccelerometer()
        {
            maxDiffAccel.X = diffAccel.X = lastAccel.X = nowAccel.X = 0.0;
            maxDiffAccel.Y = diffAccel.Y = lastAccel.Y = nowAccel.Y = 0.0;
            maxDiffAccel.Z = diffAccel.Z = lastAccel.Z = nowAccel.Y = 0.0;

            accelerometer.RequestSingleMeasurement();
            accelerometer.Calibrate();
            accelerometer.MeasurementComplete += Accelerometer_MeasurementComplete;
            accelerometer.MeasurementInterval = new TimeSpan(0, 0, 1); // every 1 sec
            accelerometer.StartTakingMeasurements();
        }
        private void SetupMobileService()
        {
            mobileService = new MobileServiceClient(new Uri(WAMSUrl), WAMSAppKey);
        }

        void MeasurementChecker(GT.Timer timer)
        {
            double twiceG248 = g248Threshold * 2;
            double twiceAccel = accelThreshold * 2;
 
            var reading = new SensorReading()
            {
                SensorDeviceId = SensorDeviceId.ToString(),
                MeasuredDate = DateTime.Now,
                MaxG248 = 0.0,
                MaxAccel = 0.0,
                Alert = 0
            };

            if (reading.MaxG248 < maxDiffG248.X)
                reading.MaxG248 = maxDiffG248.X;
            if (reading.MaxG248 < maxDiffG248.Y)
                reading.MaxG248 = maxDiffG248.Y;
            if (reading.MaxG248 < maxDiffG248.Z)
                reading.MaxG248 = maxDiffG248.Z;
            if (reading.MaxAccel < maxDiffAccel.X)
                reading.MaxAccel = maxDiffAccel.X;
            if (reading.MaxAccel < maxDiffAccel.Y)
                reading.MaxAccel = maxDiffAccel.Y;
            if (reading.MaxAccel < maxDiffAccel.Z)
                reading.MaxAccel = maxDiffAccel.Z;

            if (reading.MaxG248 > twiceG248)
            {
                reading.Alert = 2;
            }
            else if (reading.MaxAccel > twiceAccel)
            {
                reading.Alert = 2;
            }
            else if (reading.MaxG248 > g248Threshold)
            {
                reading.Alert = 1;
            }
            else if (reading.MaxAccel > accelThreshold)
            {
                reading.Alert = 1;
            }

            BeginInvoke(
                    new DispatcherOperationCallback(LedControl), reading.Alert.ToString());

            if (debug)
            {
                Debug.Print("Date:" + reading.MeasuredDate + ", "
                    + maxDiffG248.X.ToString("F2") + ", "
                    + maxDiffG248.Y.ToString("F2") + ", "
                    + maxDiffG248.Z.ToString("F2") + ", "
                    + maxDiffAccel.X.ToString("F2") + ", "
                    + maxDiffAccel.Y.ToString("F2") + ", "
                    + maxDiffAccel.Z.ToString("F2") + ", "
                    + reading.Alert);
            }

            lock(this)
            {
                maxDiffG248.X = maxDiffG248.Y = maxDiffG248.Z = 0.0;
                maxDiffAccel.X = maxDiffAccel.Y = maxDiffAccel.Z = 0.0;
            }
            try
            {
                GC.WaitForPendingFinalizers();
                var response = mobileService.GetTable(TableName).Insert(reading);
                if (response == null)
                {
                    Debug.Print("Warning! No response from Mobile Service.");
                    isAzureOK = false;
                }
                else
                {
                    isAzureOK = true;
                }
                GC.WaitForPendingFinalizers();
            }
            catch (Exception ex)
            {
                Debug.Print(ex.Message);
                isAzureOK = false;
            }

            if (debug)
            {
                Debug.Print("Azure status:" + isAzureOK);
            }

            if (!isAzureOK)
            {
                multicolorLED.TurnWhite();
            }
        }

        void Analyzer(ref AccelG248.Acceleration now,
            ref AccelG248.Acceleration last,
            ref AccelG248.Acceleration diff,
            ref AccelG248.Acceleration max)
        {
            diff.X = last.X - now.X;
            diff.Y = last.Y - now.Y;
            diff.Z = last.Z - now.Z;
            if (diff.X < 0.0)
                diff.X = (-diff.X);
            if (diff.Y < 0.0)
                diff.Y = (-diff.Y);
            if (diff.Z < 0.0)
                diff.Z = (-diff.Z);
            if (max.X < diff.X)
                max.X = diff.X;
            if (max.Y < diff.Y)
                max.Y = diff.Y;
            if (max.Z < diff.Z)
                max.Z = diff.Z;
        }

        void AccelG248_Tick(GT.Timer timer)
        {
            AccelG248.Acceleration acceleration = accelG248.GetAcceleration();
            nowG248 = acceleration;

            Analyzer(ref nowG248, ref lastG248, ref diffG248, ref maxDiffG248);

            lastG248 = nowG248;
#if false
            Debug.Print("#X: " + acceleration.X.ToString("F2") + " "
              + "Y: " + acceleration.Y.ToString("F2") + " "
              + "Z: " + acceleration.Z.ToString("F2")
               );
            Debug.Print("##" + acceleration.ToString());
#endif
        }

        void Accelerometer_MeasurementComplete(Accelerometer sender,
            Accelerometer.MeasurementCompleteEventArgs acceleration)
        {
            nowAccel.X = acceleration.X;
            nowAccel.Y = acceleration.Y;
            nowAccel.Z = acceleration.Z;

            Analyzer(ref nowAccel, ref lastAccel, ref diffAccel, ref maxDiffAccel);

            lastAccel = nowAccel;
#if false
            Debug.Print("X: " + acceleration.X.ToString("F2") + " "
              + "Y: " + acceleration.Y.ToString("F2") + " "
              + "Z: " + acceleration.Z.ToString("F2")
               );
            Debug.Print("**" + acceleration.ToString());
#endif
        }

        object LedControl(object o)
        {
            string alert = o as string;
            switch (alert)
            {
                case ("1"):
                    multicolorLED.TurnRed();
                    break;
                case ("2"):
                    multicolorLED.TurnColor(GT.Color.Yellow);
                    break;
                default:
                    multicolorLED.TurnGreen();
                    break;
            }
            return null;
        }

        public class SensorReading : IMobileServiceEntity
        {
            public int Id { get; set; }
            public string SensorDeviceId { get; set; }
            public double MaxG248 { get; set; }
            public double MaxAccel { get; set; }
            public int Alert { get; set; }
            public DateTime MeasuredDate { get; set; }
        }
    }
}
