﻿using System;
using System.Collections;
using System.Threading;
using Microsoft.SPOT;
using Microsoft.SPOT.Presentation;
using Microsoft.SPOT.Presentation.Controls;
using Microsoft.SPOT.Presentation.Media;
using Microsoft.SPOT.Presentation.Shapes;
using Microsoft.SPOT.Touch;

using MSNI = Microsoft.SPOT.Net.NetworkInformation;
using Microsoft.SPOT.Net.NetworkInformation;
using Microsoft.SPOT.Time;

using Gadgeteer.Networking;
using GT = Gadgeteer;
using GTM = Gadgeteer.Modules;
using Gadgeteer.Modules.GHIElectronics;

using GHI.Usb.Host;

using Microsoft.WindowsAzure.MobileServices;
using Microsoft.Azure.Zumo.MicroFramework.Core;

namespace EnOcean
{
    public partial class Program
    {
        private int debug = 1; // control debug display

        private MobileServiceClient mobileService;
        private static readonly string WAMSAppKey = "WAMS-APP-KEY";
        private static readonly string WAMSUrl = "http://iot-kit.azure-mobile.net/";
        private static readonly string TableName = "enocean";
        private static readonly Guid SensorDeviceId = new Guid(
        // {FF994B72-069B-41C2-8850-AC523C74AA74}
            0xff994b72, 0x69b, 0x41c2, 0x88, 0x50, 0xac, 0x52, 0x3c, 0x74, 0xaa, 0x74);

        string ipAddress = "";
        bool isNetworkConnected = false;
        bool isUsbConnected = false;

        static byte[] TimeServerIPAddress = new byte[] { 133, 243, 238, 163 }; //ntp-b2.nict.jp

        //UsbSerial serialPort = null;
        //Thread serialPortThread;

        // Network
        private enum Network { Started, Completed, Error };
        private Network azureConnection;
        private AutoResetEvent azureLock = new AutoResetEvent(true);

        // EnOcean
        //readonly uint[] sIDs = { 0x0028F2B8, 0x00298FBB }; // sample switch node ID
        //readonly uint[] tIDs = { 0x0400200F, 0x04000E10 }; // sample Temp-Humid Sensor node ID
        //readonly uint[] cIDs = { 0x04100169, 0x04100178 }; // sample CO2 Sensor node ID
        readonly uint[] sIDs = null;
        readonly uint[] tIDs = { 0x0400070C }; // Temp-Humid Sensor node ID
        readonly uint[] cIDs = null;

        //bool tempHumid = true;     // Tempareture-Humid sensor available
        bool displayHumid = true;  // Display Humidity
        bool co2Sensor = false;    // CO2 Sensor available
        bool showID = true;        // display received ID

        private const double temperatureWarningThreshold = 28.0;
        private const double humidityWarningThreshold = 70.0;

        enum PacketType
        {
            Radio = 0x01,
            Response = 0x02,
            RadioSubTel = 0x03,
            Event = 0x04,
            CommonCommand = 0x05,
            SmartAckCommand = 0x06,
            RmoteManCommand = 0x07,
            RadioMessage = 0x09,
            RadioAdvanced = 0x0A
        };

        public struct EEPData
        {
            public byte porg;
            public byte func;
            public byte type;
            public ushort manID;
            public EEPData(byte p, byte f, byte t, ushort m)
            {
                porg = p;
                func = f;
                type = t;
                manID = m;
            }
        };

        private static EEPData eepTemp = new EEPData(0xA5, 0x02, 0x05, 0x0B);
        private static EEPData eepTempHum = new EEPData(0xA5, 0x04, 0x01, 0x0B);
        private static EEPData eepCO2 = new EEPData(0xA5, 0x09, 0x04, 0x0B);
        private static EEPData eepSW = new EEPData(0xF6, 0x02, 0x01, 0x0B);

        // This method is run when the mainboard is powered up or reset.   
        void ProgramStarted()
        {
            // Subscribe to USBH event.
            Controller.UsbSerialConnected += DeviceConnectedEvent;
            //BaseDevice.DisconnectedEventHandler += DisconnectedEventHandler;

            Debug.Print("=========================");
            Debug.Print("Please remove USB Dongle!");
            Debug.Print("=========================");
            // Use Debug.Print to show messages in Visual Studio's "Output" window during debugging.
            Thread.Sleep(200); // wait for Display drawing ...

            azureConnection = Network.Started;
            SetupNetwork();
            if (isNetworkConnected)
            {
                SyncTimeService(); // Get time of day
                Thread.Sleep(500); // Wait for time event
                if (timeSynced)
                {
                    Debug.Print("Time synced: " + DateTime.Now);
                }
                else
                {
                    Debug.Print("Warning! Time is not synced");
                    multicolorLED.FadeRepeatedly(GT.Color.Yellow);
                }
            }
            else
            {
                Debug.Print("No network connected. Running on local mode.");
            }
            if (isNetworkConnected)
            {
                SetupMobileService();
            }

            Debug.Print("Program Started with " + (isNetworkConnected ? "online" : "offline"));

            Thread usbThread = new Thread(ConnectUSB);
            usbThread.Start();
        }

        //
        // Network
        //
        private void SetupNetwork()
        {
            string saveIpAddress = "";
            string saveSubnetMask = "";
            string saveGatewayAddress = "";
            string[] saveDnsAddresses = null;
            GHI.Networking.EthernetBuiltIn ei = ethernetJ11D.NetworkInterface;
            NetworkInterface en = ethernetJ11D.NetworkSettings;

            Thread.Sleep(1);
            if (!ei.Opened)
            {
                Debug.Print("now open J11D");
                ei.Open();
            }
            if (ei.IsDhcpEnabled)
            {
                Debug.Print("now enable DHCP");
                ei.EnableDhcp();
            }
            else
            {
                // Save Flash ROM settings
                char[] caIpAddress = en.IPAddress.ToCharArray();
                saveIpAddress = new string(caIpAddress);

                char[] caSubnetMask = en.SubnetMask.ToCharArray();
                saveSubnetMask = new string(caSubnetMask);

                char[] caGatewayAddress = en.GatewayAddress.ToCharArray();
                saveGatewayAddress = new string(caGatewayAddress);

                int numOfDns = en.DnsAddresses.Length;
                saveDnsAddresses = new string[numOfDns];
                if (numOfDns > 0)
                {
                    char[] dnsAddress0 = en.DnsAddresses[0].ToCharArray();
                    saveDnsAddresses[0] = new string(dnsAddress0);
                }
                if (numOfDns > 1)
                {
                    char[] dnsAddress1 = en.DnsAddresses[1].ToCharArray();
                    saveDnsAddresses[1] = new string(dnsAddress1);
                }
            }

            PrintNetworkInfo("Start settings");
            foreach (var ni in NetworkInterface.GetAllNetworkInterfaces())
            {
                if (ni.NetworkInterfaceType == NetworkInterfaceType.Ethernet)
                {
                    if (ni.IsDhcpEnabled)
                    {
                        ipAddress = ni.IPAddress;
                        if (ipAddress != null && ipAddress != "0.0.0.0")
                        {
                            break;
                        }
                        else
                        {
                            ni.RenewDhcpLease();
                            Thread.Sleep(500);
                            ipAddress = ni.IPAddress;
                        }
                    }
                    else
                    {
                        ni.EnableDhcp();
                        Thread.Sleep(500);
                        ni.EnableStaticIP(saveIpAddress, saveSubnetMask, saveGatewayAddress);
                        ni.EnableStaticDns(saveDnsAddresses);
                        ipAddress = ni.IPAddress;
                    }
                    break;
                }
            }
            isNetworkConnected = ipAddress != null && ipAddress != "0.0.0.0";
            PrintNetworkInfo("Completed");
        }

        void PrintNetworkInfo(string arg)
        {
            NetworkInterface en = ethernetJ11D.NetworkSettings;
            const string line = "-----------------";
            Debug.Print(line + " " + arg + " " + line);
            Debug.Print("IP Address:   " + en.IPAddress);
            Debug.Print("DHCP Enabled: " + en.IsDhcpEnabled);
            Debug.Print("Subnet Mask:  " + en.SubnetMask);
            Debug.Print("Gateway:      " + en.GatewayAddress);
            Debug.Print(line + line + line);
        }

        private void SyncTimeService()
        {
            TimeService.SystemTimeChanged += TimeService_SystemTimeChanged;
            TimeServiceStatus status = TimeService.UpdateNow(TimeServerIPAddress, 10);
            TimeService.SetTimeZoneOffset(540); // JST time origin 9h
        }

        bool timeSynced = false;
        void TimeService_SystemTimeChanged(object sender, SystemTimeChangedEventArgs e)
        {
            lock (this)
            {
                timeSynced = true;
            }
        }

        private void SetupMobileService()
        {
            mobileService = new MobileServiceClient(new Uri(WAMSUrl), WAMSAppKey);
        }

        object InsertMobileService(object o)
        {
            EnOceanTable queueMessage = (EnOceanTable)o;

            azureLock.Reset();

            Debug.Print("Set Lock");
            azureConnection = Network.Started;
            Debug.Print("Network Insert Start");

            try
            {
                var response = mobileService.GetTable(TableName).Insert(queueMessage);
                azureConnection = response == null ? Network.Error : Network.Completed;
                Debug.Print("Network Insert Completed OK");
            }
            catch (Exception ex)
            {
                Debug.Print(ex.Message);
                azureConnection = Network.Error;
                Debug.Print("Network Error");
            }
            azureLock.Set();
            Debug.Print("Release Lock");

            if (azureConnection != Network.Completed)
            {
                multicolorLED.TurnWhite();
            }

            return null;
        }

        void SendToAzure(uint id, int rorg, ref EEPData eep, ref byte[] data, string message)
        {
            SendToAzure(id, rorg, ref eep, ref data, message, 0, 0);
        }

        void SendToAzure(uint id, int rorg, ref EEPData eep, ref byte[] data,
            string message, double temperature, double humidity)
        {
            var reading = new EnOceanTable()
            {
                NodeId = (new Guid((int)id, (short)eep.manID, 0xEEE, eep.porg, eep.func, eep.type, 0, 0, 1, 0, 0)).ToString(),
                MeasuredDate = DateTime.Now,
                EEP = eep.porg.ToString("X2") + "-" + eep.func.ToString("X2") + "-" + eep.type.ToString("X2"),
                AD0 = 0,
                AD1 = 0,
                AD2 = 0,
                Status = 0,
                Message = message,
                Temperature = 0,
                Humidity = 0,
                Alert = 0
            };

            if (rorg == 0x20) // RPS
            {
                if (data.Length == 0)
                {
                    return;
                }
                reading.AD0 = data[0];
            }
            else if (rorg == 0x22) // 4BS
            {
                if (data.Length < 4)
                {
                    return;
                }
                reading.AD0 = data[0];
                reading.AD1 = data[1];
                reading.AD2 = data[2];
                reading.Status = data[3];
                reading.Temperature = temperature;
                reading.Humidity = humidity;

                if (reading.Temperature >= temperatureWarningThreshold)
                {
                    reading.Alert |= 1;
                }
                if (reading.Humidity >= humidityWarningThreshold)
                {
                    reading.Alert |= 2;
                }
                BeginInvoke(
                        new DispatcherOperationCallback(LedControl), reading.Alert.ToString());
            }
            else
            {
                return;
            }
            if (azureConnection != Network.Error)
            {
                //Debug.Print("GC start-1");
                //GC.WaitForPendingFinalizers();
                //Debug.Print("GC end-1");

                BeginInvoke(
                    new DispatcherOperationCallback(InsertMobileService), reading);

                Debug.Print("GC start");
                GC.WaitForPendingFinalizers();
                Debug.Print("GC end");
            }
        }

        object LedControl(object o)
        {
            string alert = o as string;
            switch (alert)
            {
                case ("1"):
                    multicolorLED.TurnRed();
                    break;
                case ("2"):
                    multicolorLED.TurnBlue();
                    break;
                case ("3"):
                    multicolorLED.TurnColor(GT.Color.Purple);
                    break;
                default:
                    multicolorLED.TurnGreen();
                    break;
            }
            return null;
        }

        public class EnOceanTable : IMobileServiceEntity
        {
            public int Id { get; set; } // Will be changed to GUID
            public string NodeId { get; set; }
            public string Message { get; set; }
            public double Temperature { get; set; }
            public double Humidity { get; set; }
            public int Alert { get; set; }
            public int AD0 { get; set; }
            public int AD1 { get; set; }
            public int AD2 { get; set; }
            public int Status { get; set; }
            public string EEP { get; set; }
            public DateTime MeasuredDate { get; set; }
        }

        //
        // Display
        //

        void DisplayDebug(string s)
        {
            Debug.Print(s);
        }
        void ConnectUSB()
        {
            Thread.Sleep(500);

            while (!isUsbConnected)
            {
                multicolorLED.TurnColor(GT.Color.Orange);
                Debug.Print("Please Connect USB Dongle!");
                Thread.Sleep(1000);
                multicolorLED.TurnOff();
                Thread.Sleep(1000);
            }
        }

        object DisplayTemperature(object obj)
        {
            string text = obj as string;
            if (text != null)
            {
                Debug.Print(text);
            }
            return null;
        }

        object DisplayHumidity(object obj)
        {
            string text = obj as string;
            if (text != null)
            {
                Debug.Print(text);
            }
            return null;
        }

        object DisplayCO2(object obj)
        {
            string text = obj as string;
            if (text != null)
            {
                Debug.Print(text);
            }
            return null;
        }

        object DisplayDebug(object obj)
        {
            string text = obj as string;
            if (text != null)
            {
                Debug.Print(text);
            }
            return null;
        }

        object labelID(object obj)
        {
            string text = obj as string;
            if (text != null)
            {
                Debug.Print(text);
            }
            return null;
        }

        object labelEEPText(object obj)
        {
            string text = obj as string;
            if (text != null)
            {
                Debug.Print(text);
            }
            return null;
        }

        void DisconnectedEventHandler(object sender, EventArgs e)
        {
            Debug.Print("Device disconnected");
        }

        void DeviceConnectedEvent(object sender, UsbSerial usbSerial)
        {
            Debug.Print("Device connected");
            isUsbConnected = true;
            Thread.Sleep(1000);
            DisplayDebug(isNetworkConnected ? "     Azure OK" : "   Local Mode");
            usbSerial.BaudRate = 57600;
            usbSerial.DataBits = 8;
            usbSerial.Parity = System.IO.Ports.Parity.None;
            usbSerial.Handshake = System.IO.Ports.Handshake.None;

            usbSerial.DataReceived += usbSerial_DataReceived;

            //serialPort = new USBH_SerialUSB(device, 57600,
            //    System.IO.Ports.Parity.None, 8,
            //    System.IO.Ports.StopBits.One);
            //serialPort.Open();

            //serialPort = usbSerial;
            //serialPortThread = new Thread(SerialPortThread);
            //serialPortThread.Start();

#if false
            switch (device.TYPE)
            {
                case USBH_DeviceType.Serial_FTDI: // FTDI connected
                    serialPort = new USBH_SerialUSB(device, 57600,
                        System.IO.Ports.Parity.None, 8,
                        System.IO.Ports.StopBits.One);
                    serialPort.Open();
                    serialPortThread = new Thread(SerialPortThread);
                    serialPortThread.Start();

                    DisplayDebug(" ");

                    break;

                default:
                    Debug.Print("Unknown device type = " + device.TYPE);
                    break;
            }
#endif
        }

        private void usbSerial_DataReceived(UsbSerial sender,
            UsbSerial.DataReceivedEventArgs e)
        {
            Debug.Print("R: " + e.Data[0].ToString("X2") + " " + e.Data.Length);
            //for (int i = 0; i < e.Data.Length; i++)
            //    Debug.Print(e.Data[i].ToString());

            //sender.Write(e.Data);
            ReceiveProcess(e.Data);
        }

        void ReceiveProcess(byte[] readBuffer)
        {
            uint options;
            uint inputValue;
            uint inputValue2;
            int dataLength = 0;
            int dataOffset = 0;
            byte optionalLength = 0;
            PacketType packetType = PacketType.CommonCommand;
            byte crc8h;
            byte crc8d;
            byte[] header = new byte[4];
            bool gotHeader = false;
            int rorg = 0;
            byte[] data = new byte[4];
            int nu = 0;
            byte[] writeBuffer = new byte[16];
            byte[] id = new byte[4];
            uint iID;
            bool validData = false;

            int i = 0;
            while (i < (readBuffer.Length - 11))
            {
                if (readBuffer[i++] != 0x55)
                {
                    Debug.Print("Unknow header mark!");
                    continue;
                }
                validData = false;

                dataLength = readBuffer[i + 0] << 8 | readBuffer[i + 1];
                optionalLength = readBuffer[i + 2];
                packetType = (PacketType)readBuffer[i + 3];
                crc8h = readBuffer[i + 4];

                gotHeader = crc8h == crc.crc8(header);
                i += 5; // Header offset
                if (debug > 1)
                    Debug.Print("Got header");

                dataOffset = i;

                if (packetType == PacketType.Radio || packetType == PacketType.RadioAdvanced)
                {
                    rorg = readBuffer[0 + dataOffset];
                    dataOffset += (rorg == 0x62 ? 2 : 0);

                    id[3] = readBuffer[1 + dataOffset];
                    id[2] = readBuffer[2 + dataOffset];
                    id[1] = readBuffer[3 + dataOffset];
                    id[0] = readBuffer[4 + dataOffset];

                    if (rorg == 0x20) // RPS
                    {
                        //dataSize = 1;
                        nu = (readBuffer[5 + dataOffset] >> 7) & 0x01;
                        data[0] = (byte)(readBuffer[5 + dataOffset] & 0x0F);
                        validData = true;
                    }
                    else if (rorg == 0x22) // 4BS
                    {
                        //dataSize = 4;
                        data[0] = readBuffer[5 + dataOffset];
                        data[1] = readBuffer[6 + dataOffset];
                        data[2] = readBuffer[7 + dataOffset];
                        data[3] = readBuffer[8 + dataOffset];
                        validData = true;
                    }
                    else if (rorg == 0x62)  // Teach-In
                    {
                        //dataSize = 4;
                        data[0] = readBuffer[5 + dataOffset];
                        data[1] = readBuffer[6 + dataOffset];
                        data[2] = readBuffer[7 + dataOffset];
                        data[3] = readBuffer[8 + dataOffset];
                        validData = true;
                    }
                    else
                    {
                        Debug.Print("Unknown rorg = " + rorg);
                    }
                }
                else if (packetType == PacketType.Response)
                {
                    Debug.Print("Response: Length:" + dataLength + " RES:" + readBuffer[0 + dataOffset]);
                    data[0] = readBuffer[0 + dataOffset]; // Return Code
                    data[1] = readBuffer[1 + dataOffset]; // APP Version[0]
                    data[2] = readBuffer[2 + dataOffset]; // APP Version[1]
                    data[3] = readBuffer[3 + dataOffset]; // APP Version[2]
                }
                else
                {
                    // We ignore other type.
                }
                if (debug > 1)
                {
                    Debug.Print("i: " + i + " offset:" + dataOffset
                        + " dlen:" + dataLength + " olen:" + optionalLength
                        + " total:" + (dataLength + optionalLength)
                        + " next:" + (dataLength + optionalLength + 1));
                }
                i += (dataLength + optionalLength + 1);

                if (!validData)
                {
                    continue;
                }

                crc8d = readBuffer[dataOffset + dataLength + optionalLength];
                byte dCRC = crc.crc8(readBuffer, dataOffset, dataLength + optionalLength);
                if (crc8d != dCRC)
                {
                    Debug.Print("Invalid data CRC!");
                    return;
                }

                // Display Data
                if (debug > 1)
                {
                    string IdData = "Data: ";
                    for (int j = 0; j < 4; j++)
                    {
                        IdData += data[j].ToString("X2");
                    }
                    Debug.Print(IdData);
                }
                //more,...
                //iID = BitConverterToInt32(id, 0);
                iID = (uint)((id[0])
                        | (id[1] << 8)
                        | (id[2] << 16)
                        | (id[3] << 24));

                if (iID != 0)
                {
                    options = (uint)data[3]; // Read Input Value from data[3] //
                    inputValue = (uint)data[2]; // Read Input Value from data[2] //
                    inputValue2 = (uint)data[1]; // Read Input Value from data[1] //

                    if (debug > 1 || showID)
                    {
                        string IdText = "ID: ";
                        IdText += iID.ToString("X8");
                        if (debug > 1)
                        {
                            string s = IdText + " " + inputValue.ToString("X2")
                                + " " + inputValue2.ToString("X2") + "\r\n";
                            Debug.Print(s);
                        }
                        if (showID)
                        {
                            DisplayID(IdText);
                        }
                    }

                    //
                    // Check and Dispatch by ID and rorg
                    //
                    if (rorg == 0x62) // Teach-In
                    {
                        // get EEP, Teach In Telegram
                        uint func = ((uint)data[0]) >> 2;
                        uint type = ((uint)data[0] & 0x03) << 5 | ((uint)data[1]) >> 3;
                        int manID = (data[1] & 0x07) << 8 | data[2];

                        if (debug > 2)
                            Debug.Print("Teach In: A5-" +
                                func.ToString("X2") + "-" +
                                type.ToString("X2") + " " +
                                manID.ToString("X3"));
                        if (manID != 0)
                            DisplayEEP((int)func, (int)type, manID);
                    }
                    else if (rorg == 0x22 &&
                        (tIDs == null || tIDs.Length > 0 && Array.IndexOf(tIDs, iID) >= 0))
                    {
                        if (debug > 2)
                            Debug.Print("4BS Temp-Humid");

                        if (displayHumid)
                        {
                            double humidity = (inputValue2 - 5) * 100 / 250;
                            if (humidity < 0.0 || humidity > 100.0)
                                humidity = 0.0;
                            string humid = humidity.ToString("F1") + " %";

                            BeginInvoke(
                                new DispatcherOperationCallback(DisplayHumidity),
                                humid);

                            double temperature = ((inputValue * 40 + 125) / 250);
                            if (temperature < 0.0)
                                temperature = 0.0;
                            string temp = temperature.ToString("F1") + "\u00B0C";

                            BeginInvoke(
                                new DispatcherOperationCallback(DisplayTemperature),
                                temp);

                            SendToAzure(iID, rorg, ref eepTempHum, ref data, temp + " " + humid,
                                temperature, humidity);
                        }
                        else
                        {
                            //DisplayTemperature(inputValue);
                            double value = ((-40 * inputValue) / 255) + 40;
                            if (value < 0.0)
                                value = 0.0;
                            string temp = value.ToString("F1") + "\u00B0C";

                            BeginInvoke(
                                new DispatcherOperationCallback(DisplayTemperature),
                                temp);
                            SendToAzure(iID, rorg, ref eepTemp, ref data, temp, value, 0);
                        }
                    }
                    else if (rorg == 0x22 &&
                        (cIDs == null || cIDs.Length > 0 && Array.IndexOf(cIDs, iID) >= 0))
                    {
                        if (debug > 2)
                            Debug.Print("4BS CO2");

                        if (co2Sensor)
                        {
                            //DisplayCO2(inputValue2);
                            int value = (int)(inputValue2) * 10;
                            string co2 = value.ToString() + " ppm";

                            BeginInvoke(
                                new DispatcherOperationCallback(DisplayCO2),
                                co2);

                            SendToAzure(iID, rorg, ref eepCO2, ref data, co2, (double) value, 0);
                        }
                    }
                    else if (rorg == 0x20 &&
                        (sIDs == null || sIDs.Length > 0 && Array.IndexOf(sIDs, iID) >= 0)) // RPS
                    {
                        if (debug > 2)
                            Debug.Print("RPS Telegram, nu=" + nu);

                        if (nu == 0x01)
                        {
                            uint mask = 0x01;
                            for (int j = 0; j < 4; j++)
                            {
                                string s = null;
                                bool sendToAzure = false;
                                switch (data[0] & mask)
                                {
                                    case 0x01:
                                        s = "Left On";
                                        sendToAzure = true;
                                        break;
                                    case 0x02:
                                        s = "Left Off";
                                        sendToAzure = true;
                                        break;
                                    case 0x04:
                                        s = "Right On";
                                        sendToAzure = true;
                                        break;
                                    case 0x08:
                                        s = "Right Off";
                                        sendToAzure = true;
                                        break;
                                } //switch
                                if (sendToAzure)
                                {
                                    SendToAzure(iID, rorg, ref eepSW, ref data, s, (double)data[0], (double) 0);
                                }
                                mask <<= 1;
                            } // for
                        } // if
                    } //else if
                    else
                    {
                        if (debug > 3)
                            Debug.Print("Unknown type");
                    }
                    if (azureConnection == Network.Error)
                    {
                        Debug.Print("Network Error!");
                        BeginInvoke(
                            new DispatcherOperationCallback(
                            DisplayDebug), "Network Error!");
                    }
                }
                else
                {
                    if (debug > 2)
                        Debug.Print("Zero ID");
                }
            }
            if (debug > 2)
                Debug.Print("Exit");
        }

        static readonly string[] hexes = new string[] { "0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "A", "B", "C", "D", "E", "F", };
        public static string X2String(byte b)
        {
            string s = hexes[(b >> 4)];
            s += hexes[b & 0x0F];
            return s;
        }

        public static string X3String(int i)
        {
            string s = hexes[((i & 0xF00) >> 8)];
            s += hexes[((i & 0xF0) >> 4)];
            s += hexes[i & 0x0F];
            return s;
        }

        public static string XnString(ref byte[] b, int length)
        {
            string s = "";
            for (int i = 0; i < length; i++)
            {
                s += X2String(b[i]);
            }
            return s;
        }

        void DisplayEEP(int func, int type, int manID)
        {
            string eepText = X2String(0xA5) + "-" + X2String((byte)func) + "-"
            + X2String((byte)type) + " " + X3String(manID);

            Debug.Print(eepText);
        }

        void DisplayID(string rId)
        {
            Debug.Print(rId);
        }
    }

    public static class crc
    {
        private static Byte[] crc8Table = new Byte[]
        {
            0x00, 0x07, 0x0e, 0x09, 0x1c, 0x1b, 0x12, 0x15,
            0x38, 0x3f, 0x36, 0x31, 0x24, 0x23, 0x2a, 0x2d,
            0x70, 0x77, 0x7e, 0x79, 0x6c, 0x6b, 0x62, 0x65,
            0x48, 0x4f, 0x46, 0x41, 0x54, 0x53, 0x5a, 0x5d,
            0xe0, 0xe7, 0xee, 0xe9, 0xfc, 0xfb, 0xf2, 0xf5,
            0xd8, 0xdf, 0xd6, 0xd1, 0xc4, 0xc3, 0xca, 0xcd,
            0x90, 0x97, 0x9e, 0x99, 0x8c, 0x8b, 0x82, 0x85,
            0xa8, 0xaf, 0xa6, 0xa1, 0xb4, 0xb3, 0xba, 0xbd,
            0xc7, 0xc0, 0xc9, 0xce, 0xdb, 0xdc, 0xd5, 0xd2,
            0xff, 0xf8, 0xf1, 0xf6, 0xe3, 0xe4, 0xed, 0xea,
            0xb7, 0xb0, 0xb9, 0xbe, 0xab, 0xac, 0xa5, 0xa2,
            0x8f, 0x88, 0x81, 0x86, 0x93, 0x94, 0x9d, 0x9a,
            0x27, 0x20, 0x29, 0x2e, 0x3b, 0x3c, 0x35, 0x32,
            0x1f, 0x18, 0x11, 0x16, 0x03, 0x04, 0x0d, 0x0a,
            0x57, 0x50, 0x59, 0x5e, 0x4b, 0x4c, 0x45, 0x42,
            0x6f, 0x68, 0x61, 0x66, 0x73, 0x74, 0x7d, 0x7a,
            0x89, 0x8e, 0x87, 0x80, 0x95, 0x92, 0x9b, 0x9c,
            0xb1, 0xb6, 0xbf, 0xb8, 0xad, 0xaa, 0xa3, 0xa4,
            0xf9, 0xfe, 0xf7, 0xf0, 0xe5, 0xe2, 0xeb, 0xec,
            0xc1, 0xc6, 0xcf, 0xc8, 0xdd, 0xda, 0xd3, 0xd4,
            0x69, 0x6e, 0x67, 0x60, 0x75, 0x72, 0x7b, 0x7c,
            0x51, 0x56, 0x5f, 0x58, 0x4d, 0x4a, 0x43, 0x44,
            0x19, 0x1e, 0x17, 0x10, 0x05, 0x02, 0x0b, 0x0c,
            0x21, 0x26, 0x2f, 0x28, 0x3d, 0x3a, 0x33, 0x34,
            0x4e, 0x49, 0x40, 0x47, 0x52, 0x55, 0x5c, 0x5b,
            0x76, 0x71, 0x78, 0x7f, 0x6A, 0x6d, 0x64, 0x63,
            0x3e, 0x39, 0x30, 0x37, 0x22, 0x25, 0x2c, 0x2b,
            0x06, 0x01, 0x08, 0x0f, 0x1a, 0x1d, 0x14, 0x13,
            0xae, 0xa9, 0xa0, 0xa7, 0xb2, 0xb5, 0xbc, 0xbb,
            0x96, 0x91, 0x98, 0x9f, 0x8a, 0x8D, 0x84, 0x83,
            0xde, 0xd9, 0xd0, 0xd7, 0xc2, 0xc5, 0xcc, 0xcb,
            0xe6, 0xe1, 0xe8, 0xef, 0xfa, 0xfd, 0xf4, 0xf3
        };

        public static Byte crc8(Byte[] data, int offset, int count)
        {
            Byte crc = 0;
            count += offset;
            for (int i = offset; i < count; i++)
                crc = crc8Table[crc ^ data[i]];
            return crc;
        }

        public static Byte crc8(Byte[] data, int count)
        {
            return crc8(data, 0, count);
        }

        public static Byte crc8(Byte[] data)
        {
            return crc8(data, data.Length);
        }
    }
}
